package model.ontology;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

/**
 * This class represents the ontology of synsets in a annotated cluster 
 * @author gonta
 *
 */
public class Ontology {
	
	private static Ontology instance;
	
	public List<Node> nodes;
	public List<Relation> edges;
	
	/**
	 * Constructor (It's private because the class will be singleton)
	 */
    private Ontology() {
    	nodes = new ArrayList<Node>();
		edges = new ArrayList<Relation>();
		nodes.add(new Node("root", 0, 0));		
		nodes.add(new Node("Verbo", 0, 0));
		nodes.add(new Node("Substantivo", 0, 0));
		Relation a1 = new Relation(nodes.get(0), nodes.get(1));
		Relation a2 = new Relation(nodes.get(0), nodes.get(2));
		edges.add(a1);
		edges.add(a2);
		nodes.get(0).addAdjacent(a1);
		nodes.get(0).addAdjacent(a2);
    }
    
    /**
     * This method returns a singleton object of Ontology
     * @return
     */
    public static Ontology getInstance(){
          if (instance == null) {
              instance = new Ontology();
          }
          return instance;
    }	
    
    /**
     * This method restart the singleton object
     */
    public static void restart(){
        instance = new Ontology();
    }

    /**
     * This method adds a node to the ontology
     * @param nome the name of the node
     * @param id id of the node
     * @param isSelected if the node added was selected
     * @return
     */
    public Node addNode(String nome, int id, boolean isSelected) {
    	Node aux = new Node (nome,id);
    	int exist = getIndex(aux);
    	if( isSelected ) {
    		if( exist == -1 ) {
    			aux.setnSelect(1);
    			aux.setCountSub(1);
    			nodes.add( aux );
    		} else {
    			nodes.get(exist).setnSelect( nodes.get(exist).getnSelect() + 1 );
    			nodes.get(exist).setCountSub( nodes.get(exist).getCountSub() + 1 );
    			aux = nodes.get(exist);
    		}
    		return aux;
    	} else {
    		if( exist == -1) {
    			nodes.add( aux );
    			return aux;
    		} else 
    			return nodes.get(exist);
    	}
    }

    /**
     * This method search a node in the ontology and returns this  if exists
     * @param aux node to found
     * @return
     */
    public Node get(Node aux) {
    	for (Node no : this.nodes) {
			if (no.equals(aux))
				return no;
		}
		return null;
	}
    
    /**
     * This method search a node in the ontology and returns the position in the list of nodes
     * @param aux node to found
     * @return
     */
    private int getIndex(Node aux) {
    	int i=0;
    	for (Node no : this.nodes) {
			if (no.equals(aux))
				return i;
			i++;
		}
		return -1;
	}    

    /**
     * This method obtains all of nodes in the ontology
     * @return
     */
	public List<Node> getNodes() {
		return nodes;
	}

	/**
	 * This method sets a list of nodes into the ontology
	 * @param nodes
	 */
	public void setNodes(List<Node> nodes) {
		this.nodes = nodes;
	}

	/**
	 * This method creates a relations between a source node and a destination node
	 * @param origem source node
	 * @param destino destination node 
	 * @param isSelected if the destination was selected (annotated) 
	 * @return
	 */
    public Relation addRelation(Node origem, Node destino, boolean isSelected) {
        Relation e = new Relation(origem, destino);
        Relation exist = origem.findEdge(destino);
        if( exist == null ){        	
            origem.addAdjacent(e);
            origem.setCountSub(origem.getCountSub()+destino.getCountSub());
            edges.add(e);	 
        } else {
        	e = exist;
        	if( isSelected ) {
        		origem.setCountSub( origem.getCountSub() + 1 );
        		e.setDestination(destino);
        		origem.addEdge(e);
        	}
        }	
        return e;
    }
		
    /**
     * This method deletes a node from the ontology
     * @param noForDelete node to be deleted
     * @return
     */
	public boolean deleteNode(Node noForDelete) {
		Node forDeleting = null; 
		Node no = get(noForDelete);
		if( no == null ) 
		   return false;
		else {
			int countSub = no.getCountSub();
			countSub--;
			no.setCountSub(countSub);
			if( no.getCountSub() == 0 ) {
				forDeleting = no;
			} else {
				int nroSelect = no.getnSelect();
				nroSelect--;
				no.setnSelect(nroSelect);
			}
		}
		
		if( forDeleting != null ) {

			this.nodes.remove(forDeleting);
			//deleting children
			List<Relation> edges = forDeleting.getAdj();
			for (Relation aresta : edges) {
				Relation auxAresta = get( aresta );
				if ( auxAresta != null ) {				
					this.nodes.remove(auxAresta.getDestination());
					this.edges.remove(auxAresta);
				}					
			}
			
			List<Relation> edgesPai = getRelationsByDestination(forDeleting); 
			for( Relation ar: edgesPai ) {
				this.edges.remove(ar);
				Node noPai = ar.getSource();
				noPai.getAdj().remove(ar);
				deleteNodeRecursive(noPai);
			}
		} else {
			List<Relation> edgesPai = getRelationsByDestination(no); 
			for( Relation ar: edgesPai ) {
				Node noPai = ar.getSource();
				deleteNodeRecursive(noPai);
			}
		}
		return true;
	}
	
	/**
	 * This method deletes the node, the relations and the hierarchy of the node
	 * @param no root node
	 * @return
	 */
	public boolean deleteNodeRecursive( Node no ) {		
		int countSub = no.getCountSub();
		countSub--;
		no.setCountSub(countSub);
		if( no.getCountSub() == 0 ) {
			List<Relation> arestaIrmaos = no.getAdj();
			for (Relation aresta : arestaIrmaos) {
				Node irmao = aresta.getDestination();
				if( irmao.getCountSub() == 0 ) 
					this.nodes.remove(irmao);
				this.edges.remove(aresta);
			}
			if( !no.equals(this.nodes.get(0)) && !no.equals(this.nodes.get(1)) && !no.equals(this.nodes.get(2)) ) {
				this.nodes.remove(no); 
				List<Relation> edgesPai = getRelationsByDestination(no);
				for( Relation ar: edgesPai ) {
					this.edges.remove(ar);
					Node noPai = ar.getSource();
					noPai.getAdj().remove(ar);
					deleteNodeRecursive(noPai);
				}	
			}
		
		} else {
			List<Relation> edgesPai = getRelationsByDestination(no); 
			for( Relation ar: edgesPai ) {
				Node noPai = ar.getSource();
				deleteNodeRecursive(noPai);
			}
		}
		return false;
	}
	
	/**
	 * This method gets the relations that have got as hyponymous a destination node
	 * @param destino destination node
	 * @return
	 */
	public List<Relation> getRelationsByDestination(Node destino) {
    	List<Relation> edges = new ArrayList<Relation>();
    	for ( Relation aresta : this.edges ) {
    		if( aresta.getDestination().equals(destino) )
    			edges.add(aresta);
    	}
		return edges;
	}

	/**
	 * This method search a relation in the ontology and returns this  if exists
	 * @param e the relation to be found
	 * @return
	 */
	private Relation get(Relation e) {
    	for (Relation aresta : edges) {
			if(aresta.equals(e))
				return aresta;
		}
		return null;
	}

	/**
	 * This method converts the ontology to a String 
	 */
	@Override
	public String toString() {
        String r = "";
        for (Node u : nodes) {
            r += u.name + "(countSub :" +u.countSub + ",nselect : " +u.nSelect+ ") -> ";
            for (Relation e : u.adj) {
                Node v = e.destination;
                r += v.name + "(countSub :" +v.countSub + ",nselect : " +v.nSelect+ ")  ";
            }
            r += "\n";
        }
        return r;
    }
	
	/**
	 * This method ....
	 * @param u
	 * @return
	 */
	public String getSubTree(Node u) {
		 String r = "";
		if(u.adj != null && u.adj.size() != 0) {
			for (Relation e : u.adj) {
				Node v = e.destination;
				r += u.name + "(nselect:" +u.nSelect+ ") \t \t " + v.name + "(nselect:" +v.nSelect+ ")\n";
				 r += getSubTree(v);
			}
		}
		return r;
	}
	
	/**
	 * This method...
	 * @return
	 */
	public String getSubTreeVerb() {
		Node nodeVerb = nodes.get(1);
		System.out.println(nodeVerb.name);
		String r = getSubTree(nodeVerb);
		return r;
    }

	/**
	 * This method adds a relation of verb or noun to a root node
	 * @param subPai node (verb or noun)
	 * @param type 1:verb 2:noun
	 */
	public void addRelationRoot(Node subPai,int type) {
		if( get(subPai) != null)
			addRelation(nodes.get(type), subPai, true);
	}
		
	/**
	 * This method saves the ontology in xml format
	 * @param path file where it must be saved
	 */
	public void saveAll(final File path){
    	 new Thread("save ontology All tmp"){
    		 @Override
			public void run() {
    			 Document document = new Document();
    	    	 Element root = new Element("root");
    			 document.setRootElement( root );
    			 
    			 Node raiz = nodes.get(0);
    			 List<Element> nos = getElementsAll(raiz.getAdj());
    			 root.addContent(nos);  	 
    	    	 File p = new File( path +	"_ontology_all"+ ".xml");
    	    	 
    	    	 XMLOutputter out = new XMLOutputter();					
    			try {
    				out.output( document, new FileOutputStream( p ));    				
    			} catch (FileNotFoundException e) {
    				e.printStackTrace();
    			} catch (IOException e) {
    				e.printStackTrace();
    			}
    		 }
    	 }.start();
     }

	/**
	 * This method returns a list of destination nodes from a list of relations
	 * @param adj list of relations for a node
	 * @return
	 */
    private List<Element> getElementsAll(List<Relation> adj) {
    	List<Element> lista = new ArrayList<Element>();   	
    	if(adj == null || adj.size() ==0){
    		return null;
    	} else {    		
    		for (Relation aresta : adj) {
    			Element element = new Element("synset");
    			element.setAttribute("name", aresta.getDestination().getName());
    			element.setAttribute("id", String.valueOf(aresta.getDestination().getId()));
    			element.setAttribute("nSelect",String.valueOf(aresta.getDestination().getnSelect()));
    			element.setAttribute("countSub",String.valueOf(aresta.getDestination().getCountSub()));
    			element.addContent(getElementsAll(aresta.getDestination().getAdj()));
    			lista.add(element);
    		}
    		return lista;
    	}    	
	}
    
    /**
     * This method loads a ontology from a file previouslly saved
     * @param origem file where the ontology is
     */
    public void load(File origem) {
    	SAXBuilder builder = new SAXBuilder();
        try {
			Document document = builder.build( origem );
			Element root = document.getRootElement();
			List<Element> childrenCG = root.getChildren();
			int i=1;
			for (Element classGramatical : childrenCG) {
				List<Element> children = classGramatical.getChildren();	 
				this.nodes.get(i).setCountSub( Integer.parseInt(classGramatical.getAttributeValue("countSub")) );
				for (Element element : children) {
					Node no = new Node(element.getAttributeValue("name"), Integer.parseInt(element.getAttributeValue("id")), Integer.parseInt(element.getAttributeValue("nSelect")), Integer.parseInt(element.getAttributeValue("countSub")) );
					this.nodes.add(no);			
					no.setAdj(processChildren(element,no));
					Relation aresta = new Relation(this.nodes.get(i),no);
					this.nodes.get(i).addAdjacent(aresta);
					this.edges.add(aresta);
				}	
				i++;
			}	
		} catch (JDOMException | IOException e) {
			e.printStackTrace();
		}
    }

    /**
     * This method processes the elements of a ontology to be loaded
     * @param element node
     * @param noP dad node
     * @return
     */
	private List<Relation> processChildren(Element element,Node noP) {
		List<Element> children = element.getChildren();
		for (Element child : children) {
			Node no = new Node(child.getAttributeValue("name"), Integer.parseInt(child.getAttributeValue("id")), Integer.parseInt(child.getAttributeValue("nSelect")), Integer.parseInt(child.getAttributeValue("countSub")) );
			Node aux = get(no); 
			if( aux == null )
				this.nodes.add(no);
			aux = no;
			aux.setAdj(processChildren(child,aux));		
			Relation aresta = new Relation(noP,aux);
			noP.addAdjacent(aresta);
			if( get(aresta) == null)
				this.edges.add(aresta);
		}
		return noP.getAdj();
	}

	public static void main(String[] args) {
		Ontology g = Ontology.getInstance();
		g.load(new File("C:\\Users\\HP\\Dropbox\\Proyectos_Disertacion\\NMLAS\\ontology\\C1_annotation_ontology_all.xml"));
		System.out.println(g.toString());
    }
    
}

	